/**
  *  \file client/scr-mine.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens.h"
#include "client/wtile.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "game/global.h"
#include "game/mine.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "game/planet.h"
#include "client/tiles/objheader.h"
#include "client/tiles/minepassage.h"
#include "client/tiles/mineinfo.h"
#include "int/if/userif.h"
#include "client/widgets/objcontrol.h"
#include "ui/widgets/button.h"
#include "game/team.h"
#include "game/config.h"
#include "game/host.h"
#include "game/racename.h"
#include "ui/richdocument.h"
#include "ui/widgets/keytranslator.h"
#include "util/misc.h"

namespace {
    class WMinefieldInfoWindow : public UIWindow {
        ObjectHolder h;
        GObjectType& type;
        void init();
     public:
        WMinefieldInfoWindow(GObjectType& type);
        bool handleEvent(const UIEvent& event, bool second_pass);
    };
}

static void
showSweepInfo(const GMinefield& mf)
{
    /* Figure out fighter sweep rate */
    int frace = ::getPlayerId();
    int frate = config.FighterSweepRate(frace);
    if (frate == 0) {
        /* We cannot sweep with fighters. Find someone who can. */
        for (int i = NUM_PLAYERS; i > 0; --i) {
            if (config.FighterSweepRate(i) > 0) {
                frate = config.FighterSweepRate(i);
                frace = i;
            }
        }
    }
    if (mf.isWeb() && (!config.AllowColoniesSweepWebs() || raceId(frace) != 11)) {
        /* This field cannot be swept with fighters */
        frate = 0;
    }

    /* Figure out units and regular sweep rate */
    int32_t units = host.isPHost() ? mf.getUnitsAfterDecay(mf.getUnits()) : mf.getUnits();
    int32_t rate = mf.isWeb() ? config.WebMineSweepRate(::getPlayerId()) : config.MineSweepRate(::getPlayerId());
    if (rate <= 0)
        rate = 1;


    /* Create dialog */
    int lines = 12 + (frate != 0);

    // VBox
    //   UIRichDocument
    //   HBox
    //     Spacer, "OK", Spacer
    ObjectHolder h;
    UIWindow& win = h.add(new UIWindow(_("Mine Sweep"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
    win.setLayout(UIVBoxLayout::instance5);

    UIRichDocument& doc = h.add(new UIRichDocument(font_sizes[FONT_NORMAL].scaledBy(20, lines), 0));
    win.add(doc);

    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    g.add(h.add(new UISpacer()));
    win.add(g);
    win.add(h.add(new UIKeyTranslator()).addStop(' ', 1).addStop(SDLK_ESCAPE, 1));

    win.pack();
    ui_root->centerWidget(win);

    /* Generate the text */
    const int em = font_ems[FONT_NORMAL];
    RichDocument& d = doc.getDocument();
    d.add(string_t(format(mf.isWeb() ? _("To sweep %d web mines, use...") : _("To sweep %d mines, use..."), units)));
    d.addParagraph();
    for (int i = 1; i <= NUM_BEAMS; ++i) {
        int32_t perBeam = i*i*rate;
        d.addRight(5*em, numToString((units + perBeam-1) / perBeam));
        d.add(string_t(" " + getBeam(i).getName()));
        d.addNewline();
    }
    if (frate != 0) {
        d.addRight(5*em, numToString((units + frate-1) / frate));
        d.add(string_t(format(_(" %s fighter"), player_racenames.getAdjName(frace))));
    }
    d.finish();

    /* Do it */
    ui_root->runModal(win);
}

/************************** WMinefieldInfoWindow *************************/

WMinefieldInfoWindow::WMinefieldInfoWindow(GObjectType& type)
    : UIWindow(_("Minefield Information"), BLUE_WINDOW, COLOR_GRAY),
      type(type)
{
    init();
}

void
WMinefieldInfoWindow::init()
{
    // VBox
    //   WObjectTypeHeaderTile
    //   WMinefieldInfoTile
    //   WMinefieldPassageTile
    //   HBox
    //     Close GoTo Planet Del <--> H
    // Difference to PCC 1.x: Up/Down are in WObjectTypeHeaderTile
    // FIXME: where are S and I?
    setLayout(UIVBoxLayout::instance5);

    UIGroup& g1 = h.add(new UIGroup(UIHBoxLayout::instance5));

    g1.add(h.add(new UIButton(0, _("Close"), SDLK_ESCAPE)).addStop());
    g1.add(h.add(new UIButton(1, _("Go To"), SDLK_RETURN)).addStop());
    g1.add(h.add(new UISpacer()));

    add(h.add(new WObjectTypeHeaderTile(type)));
    add(h.add(new WMinefieldInfoTile(type)));
    add(h.add(new WMinefieldPassageTile(type)));
    add(g1);
    add(h.add(new WObjectTypeControl(type)));

    pack();
}

bool
WMinefieldInfoWindow::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key && event.key.code == 's') {
        if (GMinefield* mf = dynamic_cast<GMinefield*>(&type.getObject(type.getCurrent())))
            showSweepInfo(*mf);
        return true;
    }
    return UIWindow::handleEvent(event, second_pass);
}

void
doMinefieldScreen()
{
    if (global_minefield_type.getCount() == 0)
        return;

    WMinefieldInfoWindow win(global_minefield_type);
    ui_root->centerWidget(win);
    int result = ui_root->runModal(win);
    if (result == 1) {
        // "Go to"
        IntUserInterfaceBinding::getCurrentInstance()->doGoToChart(global_minefield_type.getMinefield().getPos());
    }
}

/** Find nearest minefield. Returns 0 if none found. */
int
findNearestMinefield(GPoint pt)
{
    int found = 0;
    long dist2 = 0;
    bool inside = false;
    const GMinefieldContainer& mfs = getDisplayedTurn().getCurrentUniverse().getMinefields();
    for (GMinefieldContainer::const_iterator i = mfs.begin(); i != mfs.end(); ++i) {
        long ndist2 = distanceSquared(i->getPos(), pt);
        bool ninside = (ndist2 <= i->getUnits());
        /* Pick this minefield, if
           - it's the first one
           - it's closer than our previous choice
           - we're inside it but not inside the previous one */
        if (found <= 0
            || ndist2 < dist2
            || (ninside && !inside))
        {
            found  = i->getId();
            dist2  = ndist2;
            inside = ninside;
        }
    }
    return found;
}
