/**
  *  \file game/objlist.cc
  *  \brief Object Lists
  *
  *  This module contains the simple parts of object list management.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/objlist.h"
#include "cpluslib/assert.h"
#include "game/global.h"
#include "game/ship.h"
#include "game/planet.h"

/******************************* GObjectId *******************************/

/** Constructor.
    \param type Object type
    \param id   Index into object type

    Unless \c type is a GObjectList: &getType() == &type, getIndex() == id
    afterwards. If \c type is a GObjectList, this will build an object Id
    referring to the referred object directly, not through the GObjectList. */
GObjectId::GObjectId(GObjectType& type, int id)
{
    if (GObjectList* objl = dynamic_cast<GObjectList*>(&type)) {
        *this = objl->getObjectId(id);
    } else {
        this->type = &type;
        this->id   = id;
    }
}

/****************************** GObjectList ******************************/

/** Construct blank object list. */
GObjectList::GObjectList()
{ }

/** Destructor. */
GObjectList::~GObjectList()
{ }

/** Add all objects from a type.
    \param t Object type to add. Must not be empty. */
void
GObjectList::addExistingObjects(GObjectType& t)
{
    for (int id = t.getNextIndex(0); id != 0; id = t.getNextIndex(id))
        if (t.isExistant(id))
            addObject(t, id);
}

/** Add single object. */
void
GObjectList::addObject(GObjectId id)
{
    ASSERT(id.isValid());
    list.push_back(id);
}

/** Add single object. */
void
GObjectList::addObject(GObjectType& type, int id)
{
    list.push_back(GObjectId(type, id));
}

/** Add all objects at a particular location. This is used for certain ship lists.
    \param univ  universe
    \param pt    location
    \param flags ls_IncludePlanet: include planet in list;
                 ls_IncludeForeignShips: also list foreign ships
                 (default: only list ships we're playing);
                 ls_SafeShipsOnly: include only safe ships (not guessed)
    \param exclude_ship do not include this ship

    \todo We need a means to say which player must safely see the ship */
void
GObjectList::addObjectsAt(GUniverse& univ, GPoint pt, int flags, int exclude_ship)
{
    pt = pt.getNormalized();
    if (flags & ls_IncludePlanet) {
        if (int pid = univ.getPlanetAt(pt))
            addObject(global_anyplanet_type, pid);
    }
    for (int sid = 1; sid <= NUM_SHIPS; ++sid) {
        GShip& s = univ.getShip(sid);
        if (sid != exclude_ship
            && s.isExistant()
            && s.getPos() == pt
            && ((flags & ls_IncludeForeignShips) || s.isPlayed())
            && (!(flags & ls_SafeShipsOnly) || s.isReliablyVisible()))
        {
            addObject(global_anyship_type, sid);
        }
    }
}

/** Add non-object.
    \param id Id number. 0=divider, otherwise regular non-object
    \param name Name */
void
GObjectList::addNonObject(int id, string_t name)
{
    addObject(notes, notes.add(id, name));
}

/** Clear list. */
void
GObjectList::clear()
{
    container_t().swap(list);
    notes.clear();
}

/** Add dividers according a function. The passed function will return
    a string for every object; a heading (divider) will be added whenever
    that string changes. For example, when the list contains
    \verbatim
       ...object...    ...func returns...
       Ship #1         Crystalline
       Ship #9         Crystalline
       Ship #3         Empire
       Ship #17        Crystalline
    \endverbatim
    the list will look like this afterwards ("==foo==" are dividers):
    \verbatim
       == Crystalline ==
       Ship #1
       Ship #9
       == Empire ==
       Ship #3
       == Crystalline ==
       Ship #17
    \endverbatim */
void
GObjectList::addDividers(string_t (*func)(GObject&))
{
    container_t    tmpList;
    GNoteContainer tmpNotes;

    string_t    current;

    /* Note that this adds GObjectId's constructed from
       notes with an index into tmpNotes, i.e. invalid GObjectId's.
       Those will become valid when we swap the notes afterwards. */
    for (container_t::iterator i = list.begin(); i != list.end(); ++i) {
        if (i->isGameObject()) {
            string_t next = func(i->getObject());
            if (next.length() && current != next) {
                current = next;
                tmpList.push_back(GObjectId(notes, tmpNotes.add(0, current)));
            }
            tmpList.push_back(*i);
        } else if (i->isValid()) {
            GNoteObject* note = dynamic_cast<GNoteObject*>(&i->getObject());
            if (note != 0 && note->getId() != 0) {
                tmpList.push_back(GObjectId(notes, tmpNotes.addCopyOf(*note)));
            }
        }
    }
    tmpList.swap(list);
    tmpNotes.swap(notes);
}

/** Remove all dividers and nonexistant objects. Note that this will
    not disallocate the dividers. */
void
GObjectList::removeDividers()
{
    /* in-place deletion */
    // FIXME: we could clear the notes here if no non-divider notes are encountered,
    // which is a common case, to save memory. However, this often doesn't matter
    // because removeDividers is followed by addDividers which will clear the notes.
    container_t::iterator o = list.begin();
    for (container_t::iterator i = o; i != list.end(); ++i) {
        if (i->isGameObject() || i->getId())
            *o = *i, ++o;
    }
    list.erase(o, list.end());
}

/** Remove nonexistant objects. */
void
GObjectList::removeNonexistant()
{
    /* in-place deletion */
    // FIXME: note that this also removes dividers. Should it?
    container_t::iterator o = list.begin();
    for (container_t::iterator i = o; i != list.end(); ++i) {
        if (!i->isExistant())
            *o = *i, ++o;
    }
    list.erase(o, list.end());
}

void
GObjectList::init()
{
    // Nothing to do here.
}

bool
GObjectList::isExistant(int index)
{
    return index > 0
        && size_type(index) <= list.size()
        && list[index-1].isExistant();
}

/** Get object by index.
    \param index [1,getNumObjects()] */
GObject&
GObjectList::getObject(int index)
{
    ASSERT(index > 0);
    ASSERT(unsigned(index) <= list.size());
    return list[index-1].getObject();
}

/** Get object reference by index.
    \param index [1,getNumObjects()] */
GObjectId
GObjectList::getObjectId(int index) const
{
    ASSERT(index > 0);
    ASSERT(unsigned(index) <= list.size());
    return list[index-1];
}

int
GObjectList::getNextIndex(int index)
{
    while (index < int(list.size())) {
        ++index;
        if (isExistant(index))
            return index;
    }
    return 0;
}

int
GObjectList::getPreviousIndex(int index)
{
    if (index == 0)
        index = list.size() + 1;
    while (index > 1) {
        --index;
        if (isExistant(index))
            return index;
    }
    return 0;
}

/** Look up object Id.
    \param id Id to find
    \return index such that getObject(return) == id. 0 if no such object. */
int
GObjectList::getIndexFor(GObjectId id) const
{
    for (size_type i = 0; i < list.size(); ++i)
        if (list[i] == id)
            return i+1;
    return 0;
}

/** Look up object.
    \param obj object to find
    \return index such that &getObject(return) == &obj. 0 if no such object. */
int
GObjectList::getIndexFor(const GObject& obj) const
{
    for (size_type i = 0; i < list.size(); ++i)
        if (&list[i].getObject() == &obj)
            return i+1;
    return 0;
}

/** Get number of game objects. Counts all (existing) objects not including
    non-game objects (dividers). */
GObjectList::size_type
GObjectList::getNumGameObjects() const
{
    size_type count = 0;
    for (size_type i = 0; i < list.size(); ++i)
        if (list[i].isGameObject())
            ++count;
    return count;
}

/** Get number of existing objects. Counts all existing objects (not including
    dividers). */
GObjectList::size_type
GObjectList::getNumExistingObjects() const
{
    size_type count = 0;
    for (size_type i = 0; i < list.size(); ++i)
        if (list[i].isExistant())
            ++count;
    return count;
}
