/**
  *  \file game/planet.h
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_PLANET_H
#define PCC_V2_GAME_PLANET_H

#include "game/struct.h"
#include "game/coord.h"
#include "game/obj.h"
#include "game/objtypes.h"
#include "game/cargo.h"
#include "io/io.h"             // mn16_t etc.
#include "game/unitscore.h"

class GShip;
class GUniverse;
class GTurnfile;
class GMessageInformation;

/// Native races.
enum TNativeRace {
    nr_None,
    nr_Humanoid,
    nr_Bovinoid,
    nr_Reptilian,
    nr_Avian,
    nr_Amorphous,
    nr_Insectoid,
    nr_Amphibian,
    nr_Ghipsoldal,
    nr_Siliconoid
};

/// Shipyard actions. Value of the shipyard_action field.
enum TShipyardAction {
    sya_None,                   ///< No operation.
    sya_Fix,                    ///< Fix a ship.
    sya_Recycle                 ///< Recycle a ship.
};

/// Planetary Structures.
enum TPlanetaryStructure {
    str_Mines,
    str_Factories,
    str_Defense,
    str_BaseDefense
};

/** Industrial Activity Ratings. Order matches that used in UTIL.DAT
    record 6 (TUtil6SensorSweep::activity), and is also used in
    GPlanet::getIndustryLevel(int). Note that older PHost versions
    report a sixth level at 150+ structures in UTILx.DAT, this is
    considered a bug and fixed in 4.1a/3.5a. */
enum TIndustryLevel {
    ind_Minimal,                ///< Minimal activity. PHost: 0-29 structures.
    ind_Light,                  ///< Light activity. PHost: 30-59 structures. Host: 0-99 structures.
    ind_Moderate,               ///< Moderate activity. PHost: 60-89 structures.
    ind_Substantial,            ///< Substantial activity. PHost: 90-119 structures.
    ind_Heavy                   ///< Heavy activity. PHost: 120+ structures. Host: 100+ structures.
};

/** Planet. This stores data of a planet and possibly a starbase. It
    is used for played and scanned planets.

    - We always know name and Id of all planets
    - A planet can exist or not:
      - if it exists (isExistant()), we know its position, but not
        necessarily its other data
      - if it does not exist, we do not know its position. We may still
        know some other data (from scanning)
    - A planet can have three levels of data:
      - if we're playing it, we have full data (hasPlanetData()); we may
        have full data for some other planets as well
      - if we've seen it somehow, we have partial data (hasAnyPlanetData()).
      - otherwise, we don't know anything about it
    - A starbase can have two levels of data:
      - if we're playing the planet, the base is either present or not,
        and if it's present (haveBaseData()), we have full data
      - if we're not playing the planet, we only know whether there is
        a base or not. */
class GPlanet : public GMapObject {
 private:
    GPlanet();
    GPlanet& operator=(const GPlanet&);

    static int getMaxStructuresFormula(int32 clans, int threshold);

    struct MapInfo {
        int id;
        GPoint position;
        bool known_to_not_exist;
        string_t name;
        int flags;
        TAutobuildSettings autobuild;

        MapInfo(int id);
        ~MapInfo();
    };

    struct BaseInfo;
    struct PlanetInfo;

 public:
    GUnitScores unit_scores;

    enum {
        FlagPlanetExists        = 1,    ///< This planet exists.
        FlagOwnShipsInOrbit     = 2,    ///< Own ship(s) in orbit.
        FlagAlliedShipsInOrbit  = 4,    ///< Allied ship(s) in orbit.
        FlagEnemyShipsInOrbit   = 8,    ///< Enemy ship(s) in orbit.
        FlagGuessedAllyInOrbit  = 16,   ///< Guessed allied ship(s) in orbit.
        FlagGuessedEnemyInOrbit = 32    ///< Guessed enemy ship(s) in orbit.
    };
 private:
    // Planet data
    PlanetInfo*  planet_info;
    BaseInfo*    base_info;
    MapInfo      map_info;

    // Source flags. These specify which players' .dat files contained
    // the PDATA/BDATA records that make up this planet. They do NOT
    // necessarily mean that we (a) play this unit and (b) know
    // everything about it. We will, however, assume that we know
    // everything about the units we play.
    GPlayerSet planet_src_flags;
    GPlayerSet base_src_flags;

    // Cargo reservation management
    GCargoArbiter cargo;

    // Data access
    PlanetInfo&     createPlanetInfo();
    const TPlanet&  getDisplayedPlanet() const;

 public:
    GPlanet(int id);
    explicit GPlanet(const GPlanet&);
    ~GPlanet();

    /* Loading and Saving */
    void addPlanetData(const TPlanet& data, GPlayerSet srcf);
    void addBaseData(const TStarbase& data, GPlayerSet srcf);
    void addHistoryData(const TDbPlanet& data);
    void addMessageInformation(const GMessageInformation& info);
    void setXY(int x, int y);
    void setName(const string_t& name);
    void setKnownToNotExist(bool value = true);
    void setAutobuild(const TAutobuildSettings& abs);

    void getPlanetData(TPlanet& dat) const;
    void getBaseData(TStarbase& bdat) const;
    void getAutobuild(TAutobuildSettings& abs) const;
    void getHistoryData(TDbPlanet& data) const;

    /* Turn access: these are in game/trn-cmd.cc */
    void applyTurnCommand(GTurnfile& trn, int index);
    void makePlanetTurnCommands(GTurnfile& trn, const GPlanet& dis) const;
    void makeBaseTurnCommands(GTurnfile& trn, const GPlanet& dis) const;

    /* GMapObject interface */
    string_t getName() const;
    string_t getFullName() const;
    int      getId() const;
    int      getOwner() const;
    bool     isExistant() const;
    GPoint   getPos() const;

    /* Accessors: Status */
    bool hasBase() const;
    int  getStatus() const;
    bool hasAnyPlanetData() const;
    bool hasPlanetData() const;
    bool hasBaseData() const;
    GPlayerSet getPlanetSrcFlags() const;
    GPlayerSet getBaseSrcFlags() const;

    int16 getHistoryTimestamp(TDbPlanetTimestamp which_one) const;

    /* Accessors: Structures */
    mp16_t getStructures(TPlanetaryStructure ps) const;
    void   setStructures(TPlanetaryStructure ps, mp16_t count);
    mp32_t getMaxStructures(TPlanetaryStructure ps) const;
    int    getMaxStructures(TPlanetaryStructure ps, int32 clans) const;

    mp16_t getIndustryLevel() const;
    int    getIndustryLevel(int mifa) const;
    void   setIndustryLevel(mp16_t level);

    /* Accessors: Colonists */
    mn16_t getColonistHappiness() const;
    void   setColonistHappiness(mn16_t happiness);
    mp16_t getColonistTax() const;
    void   setColonistTax(mp16_t tax);
    mn32_t getColonists() const;
    void   setColonists(mn32_t clans);

    bool   isOwnerKnown() const;

    /* Formulas: colonists */
    mn32_t getColonistChange(int tax, int mifa) const;
    mn32_t getColonistChange() const;
    int32  getColonistDue(int tax) const;
    int32  getColonistDueLimited(int tax, int32& rem_inc) const;
    mp32_t getMaxSupportedColonists(int player) const;
    mp32_t getMaxSupportedColonists() const
        { return getMaxSupportedColonists(getOwner()); }

    /* Accessors: Natives */
    mp16_t getNativeGovernment() const;
    void   setNativeGovernment(mp16_t gov);
    mn16_t getNativeHappiness() const;
    void   setNativeHappiness(mn16_t happiness);
    mp16_t getNativeRace() const;
    void   setNativeRace(mp16_t race);
    mp16_t getNativeTax() const;
    void   setNativeTax(mp16_t tax);
    void   setNatives(mp32_t natives);
    mp32_t getNatives() const;

    bool   isKnownToHaveNatives() const;
    void   setKnownToHaveNatives(bool known);

    /* Formulas: Natives */
    mn32_t getNativeChange() const;
    mn32_t getNativeChange(int tax, int mifa) const;
    int32  getNativeDue(int tax) const;
    int32  getNativeDueLimited(int tax, int32 rem_inc) const;
    int32  getBovinoidSupplyContribution() const;
    int32  getBovinoidSupplyContributionLimited() const;

    /* Accessors: FCode */
    string_t getFCode() const;
    void     setFCode(const string_t& str);
    bool     isFCodeKnown() const;

    /* Accessors: Base building */
    bool     isBuildingBase() const;
    void     setBuildBaseFlag(bool b);

    /* Accessors: Environment */
    mp16_t getOreDensity(int index) const;
    void   setOreDensity(int index, mp16_t amount);
    mp32_t getOreGround(int index) const;
    void   setOreGround(int index, mp32_t amount);
    mp16_t getTemperature() const;
    void   setTemperature(mp16_t temp);

    /* Formulas: Mining */
    int    getMiningCapacity(int index, int mines) const;

    /* Accessors: Cargo */
    int32 getMoney() const;
    int32 getSupplies() const;
    int32 getCargo(GCargoType type) const;
    int32 getCargoRaw(GCargoType type) const;
    GCargoArbiter&  getCargoArbiter()
        { return cargo; }
    void changeCargoRaw(GCargoType type, int32 amount);
    bool isCargoAmountKnown(GCargoType type) const;

    /* Accessors: Base specs */
    int getBaseDamage() const;

    /* Accessors: Base mission */
    int   getBaseMission() const;
    void  setBaseMission(int mission);

    /* Accessors: Shipyard */
    int   getBaseShipyardOrder() const;
    int   getBaseShipyardId() const;
    void  setBaseShipyardOrder(int action, int id);

    /* Accessors: Tech */
    int   getBaseTech(TTechLevel level) const;
    void  setBaseTech(TTechLevel level, int value);

    /* Accessors: build order. FIXME: THIS IS INCOMPLETE */
    int          getBaseBuildHull() const;
    TBuildOrder  getBaseBuildOrder() const;
    void         setBaseBuildOrder(const TBuildOrder& order);
    int          getBaseBuildOrderHullSlot() const;
    int          getShipCloningHere(const GUniverse& trn) const;

    int          getBaseEngineStore(int slot) const;
    int          getBaseBeamStore(int slot) const;
    int          getBaseHullStoreSlot(int slot) const;
    int          getBaseLauncherStore(int slot) const;
    int          getBaseAmmoStore(int slot) const;

    void         setBaseEngineStore(int slot, int amount);
    void         setBaseBeamStore(int slot, int amount);
    void         setBaseHullStoreSlot(int slot, int amount);
    void         setBaseLauncherStore(int slot, int amount);

    /* Accessors: Autobuild */
    int32        getAutobuildGoal(TPlanetaryStructure ps) const;
    void         setAutobuildGoal(TPlanetaryStructure ps, int32 value);
    int          getAutobuildSpeed(TPlanetaryStructure ps) const;
    void         setAutobuildSpeed(TPlanetaryStructure ps, int value);

    /* Administrivia */
    void internalCheck();
    void combinedCheck2(GUniverse& univ);

    void clearOrbitFlags();
    void setShipInOrbit(const GShip& s);

    static bool isClippedCoord(GPoint p);

    static int32 getNativeDue(int tax, int race, int gov, int32 pop, int owner);
    static int32 getAmorphousBreakfast(int happy);
    static int32 getBovinoidSupplyContribution(int32 pop, int owner);
};

class GPlanetType : public GIndexedObjectType {
 public:
    GPlanetType();

    bool isExistant(int index);
    GPlanet& getObject(int index);
    GPlanet& getPlanet()
        { return getObject(getCurrent()); }
};

class GAnyPlanetType : public GIndexedObjectType {
 public:
    GAnyPlanetType();

    bool isExistant(int index);
    GPlanet& getObject(int index);
    GPlanet& getPlanet()
        { return getObject(getCurrent()); }
};

class GBaseType : public GIndexedObjectType {
 public:
    GBaseType();

    bool isExistant(int index);
    GPlanet& getObject(int index);
    GPlanet& getPlanet()
        { return getObject(getCurrent()); }
};

class GPlanetTransfer : public GCargoContainer {
    GPlanet& planet;
 protected:
    int32  getCargoOnObject(GCargoType type);
    void   changeCargoOnObject(GCargoType type, int32 amount);
    void   changeReservedCargo(GCargoType type, int32 amount);
 public:
    GPlanetTransfer(GPlanet& planet);
    ~GPlanetTransfer();

    bool     canHaveCargo(GCargoType cargo_id);
    int32    getMaxCargo(GCargoType cargo_id);
    string_t getName() const;
    int32    getDisplayOffset(GCargoType type);

    GPlanet& getPlanet() { return planet; }
};



/** Get status flags (FlagXXX). These flags mainly describe what is in
    the planet's orbit. */
inline int
GPlanet::getStatus() const
{
    return map_info.flags;
}

/** Check whether we have any information about this planet.
    \return true iff we have any information, full or partial */
inline bool
GPlanet::hasAnyPlanetData() const
{
    return planet_info != 0;
}

/** Check whether we have full planet data.
    \return true iff we have full, playable data. If yes, all
    planet accessors will return a known value, and all setters
    will work.
    \todo This does not yet handle partial planets sent through PDATA.DAT,
    as PHost before 4.0/3.4d uses */
inline bool
GPlanet::hasPlanetData() const
{
    return planet_src_flags.nonempty();
}

/** True iff we have full starbase information.
    \return true iff we have full, playable data. If yes, all
    base accessors will work. */
inline bool
GPlanet::hasBaseData() const
{
    return base_src_flags.nonempty();
}

/** Get planet source flags. This is the set of players whose PDATA
    file contains a copy of this planet (usually a unit set, but may
    be larger for unowned planets). */
inline GPlayerSet
GPlanet::getPlanetSrcFlags() const
{
    return planet_src_flags;
}

/** Get base source flags. This is the set of players whose BDATA file
    contains a copy of this base (usually a unit set). */
inline GPlayerSet
GPlanet::getBaseSrcFlags() const
{
    return base_src_flags;
}

#endif
